__version = (1, 0, 5)

__version__ = version = '.'.join(map(str, __version))
__project__ = PROJECT = __name__

#from .jsonrpc import JSONRPCProtocol, JSONRPCRequest, JSONRPCResponse
#from .exceptions import *

# lint_ignore=W0611,W0401
